@@ -14,12 +14,14 @@ from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
 from trytond.tools import grouped_slice, cursor_dict
 
+from trytond.modules.sale.stock import process_sale
+from trytond.modules.purchase.stock import process_purchase
+
 
 __all__ = ['Configuration', 'ConfigurationSequence', 'ShipmentDrop', 'Move']
 
 
-class Configuration:
-    __metaclass__ = PoolMeta
+class Configuration(metaclass=PoolMeta):
     __name__ = 'stock.configuration'
 
     shipment_drop_sequence = fields.MultiValue(fields.Many2One(
@@ -43,8 +45,7 @@ class Configuration:
             'shipment_drop_sequence').default_shipment_drop_sequence()
 
 
-class ConfigurationSequence:
-    __metaclass__ = PoolMeta
+class ConfigurationSequence(metaclass=PoolMeta):
     __name__ = 'stock.configuration.sequence'
     shipment_drop_sequence = fields.Many2One(
         'ir.sequence', "Drop Shipment Sequence", required=True,
@@ -59,7 +60,7 @@ class ConfigurationSequence:
         TableHandler = backend.get('TableHandler')
         exist = TableHandler.table_exist(cls._table)
         if exist:
-            table = TableHandler(cls, module_name)
+            table = cls.__table_handler__(module_name)
             exist &= table.column_exist('shipment_drop_sequence')
 
         super(ConfigurationSequence, cls).__register__(module_name)
@@ -148,18 +149,20 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                 ],
             ],
         depends=['company'], readonly=True)
-    supplier_moves = fields.Function(fields.One2Many('stock.move', None,
-            'Supplier Moves',
-            states={
-                'readonly': Eval('state').in_(['shipped', 'done', 'cancel']),
-                },
-            depends=['state', 'supplier']), 'get_moves', 'set_moves')
-    customer_moves = fields.Function(fields.One2Many('stock.move', None,
-            'Customer Moves',
-            states={
-                'readonly': Eval('state') != 'shipped',
-                },
-            depends=['state', 'customer']), 'get_moves', 'set_moves')
+    supplier_moves = fields.One2Many('stock.move', 'shipment',
+        'Supplier Moves',
+        filter=[('to_location.type', '=', 'drop')],
+        states={
+            'readonly': Eval('state').in_(['shipped', 'done', 'cancel']),
+            },
+        depends=['state', 'supplier'])
+    customer_moves = fields.One2Many('stock.move', 'shipment',
+        'Customer Moves',
+        filter=[('from_location.type', '=', 'drop')],
+        states={
+            'readonly': Eval('state') != 'shipped',
+            },
+        depends=['state', 'customer'])
     code = fields.Char('Code', select=1, readonly=True)
     state = fields.Selection([
             ('draft', 'Draft'),
@@ -252,21 +255,26 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'cancel': {
                     'invisible': Eval('state').in_(['cancel', 'done']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': ~Eval('state').in_(['cancel', 'draft',
                             'waiting']),
                     'icon': If(Eval('state') == 'cancel',
-                        'tryton-clear', 'tryton-go-previous'),
+                        'tryton-undo', 'tryton-back'),
+                    'depends': ['state'],
                     },
                 'wait': {
                     'invisible': Eval('state') != 'draft',
+                    'depends': ['state'],
                     },
                 'ship': {
                     'invisible': Eval('state') != 'waiting',
+                    'depends': ['state'],
                     },
                 'done': {
                     'invisible': Eval('state') != 'shipped',
+                    'depends': ['state'],
                     },
                 })
         cls._error_messages.update({
@@ -320,18 +328,6 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         if to_write:
             Move.write(*to_write)
 
-    def get_moves(self, name):
-        if name == 'supplier_moves':
-            return [m.id for m in self.moves if m.to_location.type == 'drop']
-        elif name == 'customer_moves':
-            return [m.id for m in self.moves if m.from_location.type == 'drop']
-
-    @classmethod
-    def set_moves(cls, shipments, name, values):
-        if not values:
-            return
-        cls.write(shipments, {'moves': values})
-
     @classmethod
     def create(cls, vlist):
         pool = Pool()
@@ -348,44 +344,16 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def write(cls, *args):
-        pool = Pool()
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
         super(ShipmentDrop, cls).write(*args)
         cls._set_move_planned_date(sum(args[::2], []))
 
-        actions = iter(args)
-        for shipments, values in zip(actions, actions):
-            if values.get('state', '') not in ('done', 'cancel'):
-                continue
-            with Transaction().set_context(_check_access=False):
-                move_ids = [m.id for s in shipments for m in s.customer_moves]
-                sale_lines = []
-                for sub_moves in grouped_slice(move_ids):
-                    sale_lines += SaleLine.search([
-                            ('moves', 'in', list(sub_moves)),
-                            ])
-                sales = list(set(l.sale for l in sale_lines or []))
-                Sale.process(sales)
-
-                move_ids = [m.id for s in shipments for m in s.supplier_moves]
-                purchase_lines = []
-                for sub_moves in grouped_slice(move_ids):
-                    purchase_lines += PurchaseLine.search([
-                            ('moves', 'in', move_ids),
-                            ])
-                purchases = list(set(l.purchase for l in purchase_lines or []))
-                Purchase.process(purchases)
-
     @classmethod
     def copy(cls, shipments, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['moves'] = None
+        else:
+            default = default.copy()
+        default.setdefault('moves', None)
         return super(ShipmentDrop, cls).copy(shipments, default=default)
 
     @classmethod
@@ -405,6 +373,8 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('cancel')
+    @process_sale('customer_moves')
+    @process_purchase('supplier_moves')
     def cancel(cls, shipments):
         Move = Pool().get('stock.move')
         Move.cancel([m for s in shipments for m in s.supplier_moves])
@@ -440,37 +410,92 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
         for shipment in shipments:
             product_qty = defaultdict(lambda: 0)
             product_cost = defaultdict(lambda: 0)
+
+            for c_move in shipment.customer_moves:
+                if c_move.state == 'cancel':
+                    continue
+                product_qty[c_move.product] += UoM.compute_qty(
+                    c_move.uom, c_move.quantity, c_move.product.default_uom,
+                    round=False)
+
+            s_product_qty = defaultdict(lambda: 0)
             for s_move in shipment.supplier_moves:
                 if s_move.state == 'cancel':
                     continue
-                product_qty[s_move.product] += UoM.compute_qty(s_move.uom,
-                    s_move.quantity, s_move.product.default_uom, round=False)
                 if s_move.cost_price:
                     internal_quantity = Decimal(str(s_move.internal_quantity))
                     product_cost[s_move.product] += (
                         s_move.unit_price * internal_quantity)
-            for product, cost in product_cost.iteritems():
-                qty = Decimal(str(product_qty[product]))
+
+                quantity = UoM.compute_qty(
+                    s_move.uom, s_move.quantity, s_move.product.default_uom,
+                    round=False)
+                s_product_qty[s_move.product] += quantity
+                if product_qty[s_move.product]:
+                    if quantity <= product_qty[s_move.product]:
+                        product_qty[s_move.product] -= quantity
+                        continue
+                    else:
+                        out_quantity = (
+                            quantity - product_qty[s_move.product])
+                        out_quantity = UoM.compute_qty(
+                            s_move.product.default_uom, out_quantity,
+                            s_move.uom)
+                        product_qty[s_move.product] = 0
+                else:
+                    out_quantity = s_move.quantity
+
+                if not out_quantity:
+                    continue
+                unit_price = UoM.compute_price(
+                    s_move.product.default_uom, s_move.product.list_price,
+                    s_move.uom)
+                new_customer_move = shipment._get_customer_move(s_move)
+                new_customer_move.quantity = out_quantity
+                new_customer_move.unit_price = unit_price
+                to_save.append(new_customer_move)
+
+            for product, cost in product_cost.items():
+                qty = Decimal(str(s_product_qty[product]))
                 if qty:
                     product_cost[product] = (cost / qty).quantize(cost_exp)
-            for c_move in shipment.customer_moves:
-                if c_move.state == 'cancel':
+            for c_move in list(shipment.customer_moves) + to_save:
+                if c_move.id is not None and c_move.state == 'cancel':
                     continue
-                if product_qty[c_move.product] <= 0:
-                    c_move.shipment = None
-                else:
-                    move_qty = UoM.compute_qty(c_move.uom, c_move.quantity,
-                        c_move.product.default_uom, round=False)
-                    qty = min(product_qty[c_move.product], move_qty)
-                    c_move.quantity = UoM.compute_qty(
+                c_move.cost_price = product_cost[c_move.product]
+                if c_move.id is None:
+                    continue
+                if product_qty[c_move.product] > 0:
+                    exc_qty = UoM.compute_qty(
                         c_move.product.default_uom,
-                        qty, c_move.uom)
-                    product_qty[c_move.product] -= qty
-                    c_move.cost_price = product_cost[c_move.product]
+                        product_qty[c_move.product], c_move.uom)
+                    removed_qty = UoM.compute_qty(
+                        c_move.uom, min(exc_qty, c_move.quantity),
+                        c_move.product.default_uom, round=False)
+                    c_move.quantity = max(
+                        0, c_move.uom.round(c_move.quantity - exc_qty))
+                    product_qty[c_move.product] -= removed_qty
                 to_save.append(c_move)
+
         if to_save:
             Move.save(to_save)
 
+    def _get_customer_move(self, move):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        return Move(
+            from_location=move.to_location,
+            to_location=self.customer.customer_location,
+            product=move.product,
+            uom=move.uom,
+            quantity=move.quantity,
+            shipment=self,
+            planned_date=self.planned_date,
+            company=move.company,
+            currency=move.company.currency,
+            unit_price=move.unit_price,
+            )
+
     @classmethod
     @ModelView.button
     @Workflow.transition('waiting')
@@ -510,6 +535,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('shipped')
+    @process_purchase('supplier_moves')
     def ship(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -519,6 +545,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('done')
+    @process_sale('customer_moves')
     def done(cls, shipments):
         pool = Pool()
         Move = pool.get('stock.move')
@@ -529,8 +556,7 @@ class ShipmentDrop(Workflow, ModelSQL, ModelView):
                 })
 
 
-class Move:
-    __metaclass__ = PoolMeta
+class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
 
     customer_drop = fields.Function(fields.Many2One('party.party',
